/* 
   hdcod v0.3 - a Hangul/English email-folder decoder
  ====================================================

   This program 'hdcod' is personal purpose program, I programmed this just 
   for endless curiosity about 'Hangul Coding'. But I didn't understand
   the behind for 'Hangul Code'. I think this is a today's issue for our
   Korean Internet world must be solved emergently. I don't know much about
   computer and its coding method. Moreover I didn't understand Korean 
   language structure...
    
   Copyrights: 
       All the routine in this program follow the GNU Copying Policy. Please
       read the terms specified in GNU PUBLIC LICENSE. Temporarily I do not 
       include the terms here. Each routine's copyright is pertained to its
       original author described as follows:
          o ISO-2022-KR decoding - June-Yub Lee(1992.2.29)
          o Base64 decoding - Ienup Sung(1992.7.22)
          o Quoted Printable decoding and the others - Park Myeong Seok

   Note: 
       A major feature of 'hdcod' is that no explicit code information is 
       required to decode a email or text folder. That is, this program does
       not require any character set informations. Strictly speaking, a mail
       implies one of coding informations in mail header or contents - this 
       is the basic idea for this program. This does not means 'hdcod' uses
       mail header informations such as 'Quoted Printable' for QP, contents
       character set 'iso-2022-kr' for ISO-2022-KR or 'base64' for 'Base64'.
       It looks sequential characters and judges what is the encoded method.
       Thus, it can correctly decode for one of the above three encoding 
       methods. If the given email or folder is already decoded or founded 
       unknown encoding method, 'hdcod' just dump it to an output. Just a 
       race condition of 'hdcod' is the one that the folder contains control
       characters. The line containing an unknown control character will be 
       skipped for safety but not completely - if possible avoid binary file
       as an input to 'hdcod'. If you know how a folder is encoded, however,
       you can use a decoding option to prevent automatic decoding(since 
       hdcod v0.3). I added the function for restoring stripped KSC-5601
       code. Do not believe the new function, however. You can enjoy this
       feature by comparing the results from three differnet options, i.e:
          'hdcod <stripped KSC-5601> output1'    - skip restoring
          'hdcod -7 <stripped KSC-5601> output2' - force restoring
          'hdcod -a <stripped KSC-5601> output3' - depend on my heuristics
         I decided to support filter function in 'hdcod' for those who have 
       complained about file I/O. Still file I/O is effective. Thus you can
       use 'hdcod' as a filter for some MUA(mail user agent). If your 
       environment is UNIX based, however, this program might be of no use. 
       Already there exists useful de/encoding programs for general use of 
       email decoding, e.g., 'hcode', 'metamail'. But your system is not a 
       UNIX clone such as M$ Window$ series, this program is for you. Even 
       you are using a UNIX clone, this program may be useful if your system
       administrator is so ignorant in Korean mail transfer.
         The name is similar to well-known 'hcode' because of its similar 
       usages, but you can find the differences. This decoding method can't
       be generalized, however. Only useful for encoded Korean and/or English 
       email. I intentionally use hexadecimal ASCII code to emphasize some
       characters are to be considered specially - you need ASCII table at 
       hand to read the source.

   Change log:
       1996. 8. 5. hdcod v0.3
           - Added function of restoring MSB stripped KSC-5601 code.
             (automatic decoding is not allowed by default)
             o addopted from the source code posted by Jungshik Shin 
               (jshin@pantheon.cis.yale.edu) at han.comp.mail newsgroup. 
               - thanks to him for his comments.
             o special thanks to Jong Uk Kwon(jukwon@nuclina.hoseo.ac.kr)
               for his comment about adding this function.
             o this function is still incomplete due to code confliction in
               mail(Do you have any idea how to differenciate MSB stripped 
               KSC-5601 code with US-ASCII code? - if you are so genius :)
           - Can be used as a filter for some MUA(mail user agent).
           - Supports force-decode options
             (if you know the encoding method of a mail folder.) 
           - fixed bug for misclassifying Base64 code as unknown code.
           - follows rfc1522 in decoding Q-encoded header(but not strictly).
           - More improved QP decoding.
             o Special thanks to Minix(minix@soback.kornet.nm.kr) for his 
               bug report.
       1996. 6. 8. hdcod v0.2
           - NOTE: Changed copying policy(now follows GPL terms).
           - Rewrote hdcod v0.1.
           - Can use as a general Base64 and/or Quoted Printable and/or 
             ISO-2022-KR decoder.
           - Improved pattern checking methods: still buggie but almost safe. 
           - Improved mail header decoding and more stable QP decoding. 
           - Decodes sequentially per line basis.
           - Bug for decoding a multiple folder was fixed.
       1995. 5. 1. hdcod v0.1
           - Rewrote mime2han. 
           - Mixed code decoder.
           - ISO-2022-KR & Base64 decoding were added. 
           - Decoding was performed sequentially per character basis.
       1995. 4.12. mime2han 
           - The original misleading name of hdcod was written. 
           - Supported only QP decoding.

   Usage: hdcod [ -<7|a|b|i|q> ] [ input ] [ output ]

      Given a NoMSB and/or QP and/or B64 and/or ISO Korean/English 
      mixed folder, converts it to KS code.
         QP: Quoted Printable, B64: Base64, ISO: ISO-2022-KR
         NoMSB: MSB(8th bit) stripped KSC-5601, KS: KSC-5601

      input : NoMSB/QP/B64/ISO Korean/English folder(default: standard input)
      output: KSC-5601 file(default: standard output)
      option:
         (none): automatic decoding except KS-NoMSB
           -7  : force KS-NoMSB restore
           -a  : force automatic decoding with KS-NoMSB restore
           -b  : force Base64 decoding
           -i  : force ISO-2022-KR decoding
           -q  : force Quoted-Printable decoding

   Bug report or any comments to: pms@romance.kaist.ac.kr
       - please include 'diff' result, if possible: do the 3 steps below.
         1. # hdcod input output
         2. # diff -u input output > error.bug
         3. And include the 'error.bug' file in your mail of bug report.
*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#define BUF 2048
FILE *FPI = stdin, *FPO = stdout;
static char *FNI, *FNO;
typedef enum { 
    AUTODECODE, ISO2022KR, KSNMSB, QUOTEDPRINTABLE, BASE64 
} TDeMethod;
TDeMethod DECODE;
int ForceAutoDecode = 0;

static char *HelpText[] = {
  "\nhdcod v0.3: Decodes a NoMSB/QP/B64/ISO Korean/English folder",
  " to a KS folder.\n",
  "   QP: Quoted Printable, B64: Base64, ISO: ISO-2022-KR\n",
  "   NoMSB: MSB(8th bit) stripped KSC-5601, KS: KSC-5601\n",
  "       - pms@romance.kaist.ac.kr\n\n",
  "hdcod [ -<7|a|b|i|q> ] [ input ] [ output ]\n",
  "   input : NoMSB/QP/B64/ISO Korean/English folder",
  "(default: standard input)\n",
  "   output: KSC-5601 file(default: standard output)\n",
  "   option:\n",
  "      (none): automatic decoding except KS-NoMSB\n",
  "        -7  : force KS-NoMSB restore\n",
  "        -a  : force automatic decoding with KS-NoMSB restore\n",
  "        -b  : force Base64 decoding\n",
  "        -i  : force ISO-2022-KR decoding\n",
  "        -q  : force Quoted-Printable decoding\n\n",
  ""
};

void HelpMsg(void)
{
    char **m = HelpText;
    while(**m) printf(*m++);
}

void Dump(const char *s)
{
    fputs(s, FPO);
}

/*
**    Base64 decoding - Ienup Sung(1992.7.22)
*/
void DeB64(char *ob, const char *ib)
{
    register unsigned long bm=0;
    register short cnt=0;

    for(; *ib!=0x00; ib++) {
        bm = (*ib==0x2B)?  (bm << 6) | 0x3E:
             (*ib==0x2F)?  (bm << 6) | 0x3F:
             isdigit(*ib)? (bm << 6) | (0x34 + *ib - 0x30):
             isupper(*ib)? (bm << 6) | (*ib - 0x41):
             islower(*ib)? (bm << 6) | (0x1A + *ib - 0x61):
             (bm << 6) | 0;
        if(++cnt >= 4) {
            *ob++ = (char)((bm & 0x0FF0000) >> 16);
            *ob++ = (char)((bm & 0x000FF00) >> 8);
            *ob++ = (char) (bm & 0x00000FF);
            cnt = bm = 0;
        }
    }
    *ob = 0x00;
}

void DecodeB64(const char *s)
{
    char ts[BUF];

    DeB64(ts, s); Dump(ts);
}

/*
**    ISO-2022-KR decoding - June-Yub Lee(1992.2.29)
*/
void DecodeISO2022KR(const char *s)
{
    register int i, kor, sl = strlen(s);    

    for(i=kor=0; i<sl; i++) {
        if(i<sl-3 & s[i]==0x1B & s[i+1]==0x24 & s[i+2]==0x29 & s[i+3]==0x43)
            i += 3;
        else if(s[i]==0x0E) kor = 1;
        else if(s[i]==0x0F) kor = 0;
        else if(kor) {
            if(s[i]==0x20|s[i]==0x09|s[i]==0x0A|s[i]==0x0D) fputc(s[i], FPO);
            else fputc(s[i] + 0x80, FPO);
        } 
        else fputc(s[i], FPO);
    }
}

int IsBin(char c)
{
    if(c==0x1B || c==0x0E || c==0x0F || c==0x09 || c==0x0A || c==0x0D)
        return 0;
    return ((c>0x00 && c<0x20) || c==0x7F)? 1: 0;
}
 
int IsB64(char c)
{
    return 
        (isalnum(c)||c==0x0A||c==0x0D||c==0x2B||c==0x2F||c==0x3D)? 1: 0; 
}

int IsWhite(const char *s)
{
    register c;
    for(; *s; s++) {
        c = *s;
        if(c==0x09 || c==0x0A || c==0x0D || c==0x20) continue;
        else return 0;
    }
    return 1;
}

/*
**    Quoted Printable decoding
*/
int DeQP(const char *s)
{
    register int i, hc, hb;
    
    for(i=0; i<2; i++) {
        hb = (s[i] > 0x2F & s[i] < 0x3A)? s[i] - 0x30:
             (s[i] > 0x40 & s[i] < 0x47)? s[i] - 0x37:
             (s[i] > 0x60 & s[i] < 0x67)? s[i] - 0x57:
             -1;
        if(hb<0) return 0;
        hc = (i)? hc | hb: hb << 4; 
    }
    return hc;
}

/*
**  Quoted Printable decoding
*/
void DecodeQP(const char *s, int hf)
{
    char c, h[2], ns[BUF];
    register int i, j, join, sl = strlen(s);

    for(i=join=0; i<sl; i++) {
        if(s[i]==0x3D) {
            if(i==sl-2) { join = 1; break; }
            else if(i<sl-2) {
                for(j=0; j<2; j++) h[j] = s[i+1+j];
                if(!(c=DeQP(h))) fputc(0x3D, FPO);
                else { i+= 2; fputc(c, FPO); }
            }
        }
        else if(hf && s[i]==0x5F) fputc(0x20, FPO);
        else fputc(s[i], FPO);
    }
    if(join) {
        fgets(ns, BUF, FPI);
        DecodeQP(ns, hf);
    }
}

int HeaderCode(char c)
{
    return (c==0x42 | c==0x62)? 4: (c==0x51 | c==0x71)? 3: 0;
}

void DeHeader(int code, const char *s)
{
    if(code==3) DecodeQP(s, 1);
    if(code==4) DecodeB64(s);
}

/*
**  Email Header Decoding
**
**      QP=3, B64=4
*/
void DecodeHeader(const char *s)
{
    char ts[BUF];
    register int i, j, c, code, sl, fail;

    sl = strlen(s);
    for(i=code=0; i<sl; i++) {
        if(i<sl-1 & s[i]==0x3D & s[i+1]==0x3F) {
            fail = 0;
            if(i<sl-3 & s[i+3]==0x3F) {
                if(!code) code = HeaderCode(s[i+2]);
                if(code) {
                    for(j=0; j<sl-i-4 & s[i+4+j]!=0x3F; j++);
                    if(j>sl-i-4) fail = 1;
                    else c = i + 5 + j;
                } else fail = 1;
            }
            else {
                for(j=0; j<sl-i-2 & s[i+2+j]!=0x3F; j++);
                if(j<sl-i-3) {
                    if(!code) code = HeaderCode(s[i+3+j]); 
                    if(code && j<sl-i-4 & s[i+4+j]==0x3F) c = i + 5 + j;
                    else fail = 1;
                } else fail = 1;
            }
            if(!fail) {
                for(j=0; c<sl & s[c]!=0x3F; j++, c++) ts[j] = s[c];
                if(c<sl-1 & s[c+1]==0x3D) {
                    c++;
                    ts[j] = 0; DeHeader(code, ts);
                } else fail = 1;
            }
            if(fail) { 
                for(j=0; j<2; j++) fputc(s[i+j], FPO);
                i++;
            } else i = c;
        } else fputc(s[i], FPO); 
    }
}

int IsDigit(char c)
{
    return 
        (isdigit(c) || c==0x2E || c==0x3A || c==0x2F
                    || c==0x2F || c==0x3C || c==0x3F)? 1: 0;
}

int IsAlpha(char c)
{
    return 
        (isalpha(c) || c==0x2E || c==0x3A || c==0x2F || c==0x40)? 1: 0;
}

/*
**  Check stripped KSC 5601
**     - ZZerbi & ZZuab!! 
**     - can't verify exactly MSB stiripped KSC-5601 code.
**     - Do you have any idea??
*/
int IsKSNoMSB(const char *s)
{
    char c, eng=0, kor=0;
    register int i, w, wht, sl, ks, ksf, ksc, pn, isn, rlt;
    
    sl = strlen(s);
    for(i=0; i<sl; i++) if(s[i] & 0x80) return 0;
    for(i=0; i<sl && ((c=s[i])==0x09 || c==0x20); i++); w = i;
    for(i=w+2,wht=ks=ksc=isn=rlt=pn=0; i<sl; i++, isn=0) {
        if(wht) {
            while(s[i]!=0x20 && s[i]!=0x09) i++;
            wht = ksf = ks = 0;
            kor = eng = 0;
        }
        isn += IsDigit(s[i]);
        if(s[i]==0x20 || s[i]==0x09) {
            if(i!=w+2 && !ks && !eng) return 0;
            if(ksf && ks==1) rlt = 0;
            ksf = ks = 0;
        }
        else if(s[i] < 0x30 || s[i] > 0x48) {
            if(eng) kor = 0;
            if(kor) { wht = 1; kor = 0; }
            eng = 1;
        }
        else if(!s[++i]) break;
        else if(s[i]==0x20 || s[i]==0x09) i--;
        else if(s[i] < 0x21 || s[i] == 0x7F) {
            if(eng) kor = 0;
            if(kor) { wht = 1; kor = 0; }
            eng = 1;
        }
        else {
            if(kor) eng = 0;
            if(eng) { kor = 1; eng = 0; }
            if(!ks) ksf = 1;
            isn += IsDigit(s[i]);
            if(isn!=2) { 
                if(pn==2) { ks++; ksc++; }
                ks++; ksc++; rlt = 1;
            }
            else if(i<sl-1 && (s[i+1]==0x20 || s[i+1]==0x09)) {
                ks++; ksc++; rlt = 1;
            }
            else {
                if(eng) kor = 0;
                if(kor) { wht = 1; kor = 0; }
                eng = 1;
            }
            pn = isn;
            if(ks>2) return rlt;
        }
    }
    if(ksc==1) return 0;
    return rlt;
}

/*
**  Adopted from the comment posted at han.comp.mail 
**     by Jungshik Shin(jshin@pantheon.cis.yale.edu). 
*/
void DecodeKSNoMSB(const char *s)
{
    register int i, sl, isn, isa;
    char c;

    sl = strlen(s);
    for(i=isa=isn=0; i<sl; i++,isn=0) {
        c = s[i];
        isn += IsDigit(c);
        if(c < 0x30 || c > 0x48) fputc(c, FPO);
        else if(!s[++i]) return;
        else if(s[i] < 0x21 || s[i] == 0x7F) { 
            fputc(c, FPO); fputc(s[i], FPO); 
        }
        else { 
            isn += IsDigit(s[i]);
            if(isn!=2) { fputc(c+0x80, FPO); fputc(s[i]+0x80, FPO); }
            else { fputc(c, FPO); fputc(s[i], FPO); }
        }
    }
}

/*
**  Check Code Pattern: depends on my knowledge of encoded email folder.
**                      - this means my method is not standard.
**             
**      DUMP=0, BIN=1, ISO2022KR=2, QP=3, B64=4, KSNMSB=5, CODEDHEADER=6
*/
int CheckPattern(const char *s)
{
    char c;
    register int h, i, w, cqp, tqp, nqp, nb64, eb64, sl, coh, sch;
    
    for(i=0; i<(sl=strlen(s)) & ((c=s[i])==0x09 | c==0x20); i++); w = i;
    for(i=w, sch=coh=tqp=cqp=nb64=eb64=0; i<sl; i++) {
        if(IsBin(s[i])) return 1;
        if(i<sl-3 & s[i]==0x1B & s[i+1]==0x24 & s[i+2]==0x29 & s[i+3]==0x43)
            return 2;
        if(s[i]==0x0E | s[i]==0x0F) return 2;
        if(s[i]==0x3D) cqp++;
        if(i<sl-5 & s[i]==0x3D) {
            if(!tqp) tqp = ((s[i+1]==0x33) & (s[i+2]==0x44));
            nqp = isxdigit(s[i+1]) + isxdigit(s[i+2]) 
                                   + isxdigit(s[i+4]) + isxdigit(s[i+5]);
            if(nqp==4 & s[i+3]==0x3D) return 3;
        }
        if(s[i]==0x3D & s[i+1]==0x3F) coh = 1;
        for(h=i+2; coh && h<sl; h++) {
            if(s[h]==0x3F) coh++;
            if(coh > 2) return 6;
        }
        nb64 += IsB64(s[i]);
        if(i>w && !eb64 && s[i]==s[i-1]) { sch++; if(sch > 3) eb64 = 1; }
        else sch = 0;
    }
    if(!w && nb64==sl && !eb64) return 4;
    if((s[sl-2]==0x3D && cqp==1) | tqp) return 3;
    if(sl>3 && s[sl-4]==0x3D && isxdigit(s[sl-3]) && isxdigit(s[sl-2]))
        return 3;
    if(eb64) return 0;
    return (IsKSNoMSB(s))? 5: 0;
}

/*
**  DUMP=0, BIN=1, ISO2022KR=2, QP=3, B64=4, KSNOMSB=5, CODEDHEADER=6
*/
void Decode(int p, const char *s)
{
    switch(p) {
        case  0: Dump(s); break;
        case  1: break;
        case  2: DecodeISO2022KR(s); break;
        case  3: DecodeQP(s, 0); break;
        case  4: DecodeB64(s); break;
        case  5: DecodeKSNoMSB(s); break;
        case  6: DecodeHeader(s); break;
        default: break;
    }
}

/*
** Simple argument parsing 
*/
int ParseArgs(int ac, char **av)
{
    register int err, code, sfi, sfo;

    if(ac--==1) return 1;
    for(av++, err=code=sfi=sfo=0; ac && !err; ac--, av++) {
        if(**av=='-' || **av=='/') {
            if(strlen(*av) > 2) return 0;
            switch (*(*av+1)) {
                case '7':
                    DECODE = KSNMSB; code++;
                    break;
                case 'a':
                    ForceAutoDecode = 1;
                    break;
                case 'b':
                    DECODE = BASE64; code++;
                    break;
                case 'i': 
                    DECODE = ISO2022KR; code++;
                    break;
                case 'q':
                    DECODE = QUOTEDPRINTABLE; code++;
                    break;
                case 0: 
                    if(**av=='-') {
                        if(!sfi) { sfi = 1; break; }
                        if(!sfo) { sfo = 1; break; }
                    }
                    else err = 1;
                    break;
                default: 
                    err = 1;
                    break;
            }
        }
        else {
            if(sfi && sfo) return 0;
            if(!sfi) { FNI = *av; sfi = 1; continue; }
            if(!sfo) { FNO = *av; sfo = 1; continue; }
        }
    }
    if(err || code > 1) return 0;
    if(!code) DECODE = AUTODECODE;
    return 1;
}

int GetDeMethod(void)
{
    return (DECODE==ISO2022KR)? 2: (DECODE==QUOTEDPRINTABLE)? 3:
           (DECODE==BASE64)? 4: (DECODE==KSNMSB)? 5: 0; 
}

int main(char ac, char **av)
{
    char c;
    static char s[BUF];
    register int pbs, wht, pat, mtd;

    if(!ParseArgs(ac, av)) { HelpMsg(); return 0; }
    if(FNI && !(FPI=fopen(FNI, "r"))) { perror(FNI); HelpMsg(); return 0; }
    if(FNO && !(FPO=fopen(FNO, "w"))) { perror(FNO); HelpMsg(); return 0; }

    mtd = GetDeMethod();
    for(pbs=wht=pat=0; fgets(s, BUF, FPI)!=NULL; pbs=0, *s=0) {
        if(IsWhite(s)) { Dump(s); continue; }
        if(pat==4) pbs = 1; 
        pat = CheckPattern(s);
        if(pat==4 && strlen(s)<50 && !pbs) pat = 0;
        if(!ForceAutoDecode && !mtd && pat==5) pat = 0;
        if(mtd && pat!=6) pat = mtd;
        Decode(pat, s);
    }

    if(FNI) fclose(FPI); if(FNO) fclose(FPO);
    return 0;
}
